package jess;

import java.util.*;
import java.io.*;

/** **********************************************************************
 * An execution context for Funcalls.
 * <P>
 * (C) 1998 E.J. Friedman-Hill and the Sandia Corporation
 ********************************************************************** */
public class Context implements Serializable {
    private Hashtable m_variables;
    private Context m_parent;  
    private boolean m_return;
    private Value m_retval;
    private transient Rete m_rete;
    private LogicalNode m_logicalSupportNode;
    
    private Token m_token;
    private Fact m_fact;

    /**
     * If this context represents a join network node from a rule LHS,
     * this will return the left input of the node.
     * @return The Token
     */

    public final Token getToken() { return m_token; }
    final void setToken(Token t) { m_token = t; }

    /**
     * If this context represents a join network node from a rule LHS,
     * this will return the right input of the node.
     * @return The ValueVector form of a fact
     */

    public final Fact getFact() { return m_fact; }
    final void setFact(Fact f) { m_fact = f; }

    public final LogicalNode getLogicalSupportNode() {
        return m_logicalSupportNode;
    }

    final void setLogicalSupportNode(LogicalNode node) {
        m_logicalSupportNode = node;
    }

    private boolean m_inAdvice;
  
    boolean getInAdvice() {return m_inAdvice;}
    void setInAdvice(boolean  v) {m_inAdvice = v;}
  

    void setEngine(Rete r) {
        m_rete = r;
    }


    /**
     * @param c 
     */
    Context(Context c) {
        m_rete = c.m_rete;
        m_parent = c;
    }

    /**
     * @param engine 
     */
    Context(Rete engine) {
        m_rete = engine;
        m_parent = null;
    }

    Context(Context c, Rete engine) {
        m_rete = engine;
        m_parent = c;
    }
  
    /**
     * @return 
     */
    final boolean returning() {
        return m_return;
    }

    /**
     * @param val 
     * @return 
     */
    final Value setReturnValue(Value val) {
        m_return = true;
        m_retval = val;
        return val;
    }

    /**
     * @return 
     */
    final Value getReturnValue() {
        return m_retval;
    }

    final void clearReturnValue() {
        m_return = false;
        m_retval = null;
    }

    private final int nVariables() {
        if (m_variables == null)
            return 0;
        else
            return m_variables.size();
    }

    private Hashtable getVariables() {
        if (m_variables == null)
            m_variables = new Hashtable(10);

        return m_variables;
    }

    /**
     * @return 
     */
    public final Rete getEngine() {
        return m_rete; 
    }

    /**
     * @return 
     */
    Context push() {    
        return new Context(this);
    }
  
    /**
     * @return 
     */
    Context pop() {        
        if (m_parent != null) {
            synchronized (m_parent) {
                m_parent.m_return = m_return;
                m_parent.m_retval = m_retval;        
                return m_parent;
            }
        }
        else
            return this;
    }

    private Hashtable findVariable(String key) {
        Context c = this;
        while (c != null) {
            Hashtable ht = c.getVariables();
            Value v = (Value) ht.get(key);
            if (v != null)
                return ht;
            else
                c = c.m_parent;
        }
        return null;
    }
  
    synchronized void removeNonGlobals() {
        if (m_variables == null)
            return;

        Hashtable ht = new Hashtable(10);    
        for (Enumeration e = m_variables.keys(); e.hasMoreElements();) {
            String s = (String) e.nextElement();
            if (s.startsWith("*"))
                ht.put(s, m_variables.get(s));
        }
        m_variables = ht;
    }

    /**
     * Get the value of a variable
     * @param name The name of the variable with no leading '?' or '$'
     * characters
     * @exception JessException If the variable is undefined */
    public Value getVariable(String name) throws JessException {
        Hashtable ht = findVariable(name);
        if (ht == null)
            throw new JessException("Context.getVariable",
                                    "No such variable", name);
        return ((Value) ht.get(name)).resolveValue(this);
    }

    /**
     * Set a (possibly new) variable to some type and value
     * @param name Name of the variable
     * @param value The value of the variable
     */
    public void setVariable(String name, Value value) throws JessException {
        Hashtable ht = findVariable(name);
        if (ht == null)
            ht = getVariables();
        ht.put(name, value);
    }

    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("[Context, " + nVariables() + " variables: ");
        for (Enumeration e = getVariables().keys(); e.hasMoreElements(); ) {
            Object o = e.nextElement();
            sb.append(o + "=" + m_variables.get(o) + ";");
        }
        sb.append("]");
        return sb.toString();
    }

}


