package jess;

import java.io.Serializable;
import java.util.*;

/** 
 * A group on the LHS of a rule, like an AND, OR, NOT, EXPLICIT, or
 * other special CE. 
 * (C) 2001 E.J. Friedman-Hill and Sandia National Laboratories
 */

class Group implements Groupable, Serializable {
    private String m_name;
    private boolean m_unique, m_explicit, m_logical;
    private GroupableVector m_data;
    private boolean m_unary = false;
    protected boolean m_startedAsANot;
    
    static final String
        AND="and", UNIQUE="unique", EXPLICIT="explicit",
        NOT="not", EXISTS="exists", TEST="test", OR="or",
        LOGICAL="logical";
    
    private final static Pattern s_initialFactPattern =
        new Pattern(Deftemplate.getInitialTemplate().getName(),
                    Deftemplate.getInitialTemplate());

    private final static Pattern s_logicalInitialFactPattern;

    static {
        s_logicalInitialFactPattern =
            new Pattern(Deftemplate.getInitialTemplate().getName(),
                        Deftemplate.getInitialTemplate());
        s_logicalInitialFactPattern.setLogical();
    }


    Group(String name) {
        m_data = new GroupableVector();
        m_name = name;
        
        if (m_name.equals(UNIQUE)) {
            m_unique = true;
            m_unary = true;
        }
        
        else if (m_name.equals(EXPLICIT)) {
            m_explicit = true;
            m_unary = true;
        }
        
        else if (m_name.equals(LOGICAL)) {
            m_logical = true;
            m_unary = false;
        }
        
        else if (m_name.equals(NOT)) {
            m_unary = true;
            m_startedAsANot = true;
        }
        
        else if (m_name.equals(EXISTS)) {
            m_unary = true;
            m_startedAsANot = true;
        }    
    }
    
    public String getName() {
        return m_name;
    }
    
    void add(Groupable g) throws JessException {
        if (m_data.size() > 0 && m_unary)
            throw new JessException("Group.add",
                                    "CE is a unary modifier", m_name);
        
        else if (m_name.equals(LOGICAL) && g.getName().equals(TEST))
            throw new JessException("Group.add",
                                    "CE can't be used in logical:", "test");

        else if (m_name.equals(NOT) && g.getName().equals(LOGICAL))
            throw new JessException("Group.add",
                                    "CE can't be used in not:", "logical");

        else if (m_name.equals(EXISTS) && g.getName().equals(LOGICAL))
            throw new JessException("Group.add",
                                    "CE can't be used in exists:", "logical");
        
        // Remove single-branch ORs if added to anything.
        if (g.getName().equals(OR) && g.getGroupSize() == 1)
            g = g.getGroupable(0);


        // Remove single-branch ANDs if added to anything, unless they
        // are being added to an OR CE.
        if (g.getName().equals(AND) && g.getGroupSize() == 1 &&
            !getName().equals(OR))
            g = g.getGroupable(0);       
        
        
        // Flatten nested ORs immediately.
        if (m_name.equals(OR) && g.getName().equals(OR)) {
            Group grp = (Group) g;
            for (int i=0; i<grp.m_data.size(); i++)
                add(grp.m_data.get(i));
        }

        // Flatten nested ANDs immediately.
        else if (m_name.equals(AND) && g.getName().equals(AND)) {
            Group grp = (Group) g;
            for (int i=0; i<grp.m_data.size(); i++)
                add(grp.m_data.get(i));
        }
        
        // Rearrange an (OR) inside of a (NOT)
        // According to (not (or A B)) => (and (not A) (not B))
        else if (m_name.equals(NOT) && g.getName().equals(OR)) {
            m_name = AND;
            m_unary = false;
            Group grp = (Group) g;
            for (int i=0; i<grp.m_data.size(); i++) {
                Group not = new Group(NOT);
                not.add(grp.m_data.get(i));
                add(not);
            }
        }

        // Rearrange an (AND) inside of a (NOT)
        // According to (not (and A B)) =>
        // (or  (A and not B) (B and not A) (not A and not B))        
        else if (m_name.equals(NOT) && g.getName().equals(AND)) {
            m_name = OR;
            m_unary = false;

            Group beingAdded = (Group) g;
            int size = beingAdded.m_data.size();

            // We generate all the cases by modeling the group as a
            // binary number. Each digit is one pattern. We count off
            // the permutations.  When a digit in the number is set,
            // we make the corresponding pattern negated in that
            // permutation.
            int limit = (int) Math.pow(2, size);
            for (int permutation=1; permutation<limit; ++permutation) {
                Group clause = new Group(AND);
                for (int digit=0; digit < size; ++digit) {
                    if (bitIsSet(permutation, digit)) {
                        Group not = new Group(NOT);
                        not.add(beingAdded.m_data.get(digit));
                        clause.add(not);
                    } else {
                        clause.add(beingAdded.m_data.get(digit));
                    } 
                }
                add(clause);
            }
        } 

        // Otherwise keep the tree-like structure
        else
            m_data.add(g);
    
        if (m_explicit)
            setExplicit();
    
        if (m_unique)
            setUnique();    

        if (m_logical)
            setLogical();    
    }

    public boolean getBackwardChaining() {
        return false;
    }

    boolean bitIsSet(int bitfield, int index) {
        int mask = (int) Math.pow(2, index);
        return (bitfield & mask) != 0;
    }
  
    public void setExplicit() {
        for (int i=0; i<m_data.size(); i++)
            m_data.get(i).setExplicit();
        m_explicit = true;
    }

    public void setUnique() {
        for (int i=0; i<m_data.size(); i++)
            m_data.get(i).setUnique();
        m_unique = true;
    }

    public void setLogical() {
        for (int i=0; i<m_data.size(); i++)
            m_data.get(i).setLogical();
        m_logical = true;
    }

    public boolean getLogical() {
        return m_logical;
    }

    public void setBoundName(String name) throws JessException {
        if (m_name.equals(NOT) ||
            m_name.equals(TEST) ||
            (m_data.size() > 1 && !m_name.equals(OR)))
            throw new JessException("Group.setBoundName",
                                    "This CE can't be bound to a variable",
                                    m_name);

        for (int i=0; i<m_data.size(); i++)
            m_data.get(i).setBoundName(name);
    }

    static boolean isGroupName(String s) {
        return
            isNegatedName(s) ||
            s.equals(AND) ||
            s.equals(OR) ||
            s.equals(UNIQUE) ||
            s.equals(LOGICAL) ||
            s.equals(EXPLICIT);     
    }

    static boolean isNegatedName(String s) {
        return
            s.equals(NOT) ||
            s.equals(EXISTS);
            
    }

    public String toString() {
        StringBuffer sb = new StringBuffer("(");
        sb.append(m_name);
        for (int i=0; i<m_data.size(); i++) {
            sb.append('\n');
            sb.append(' ');
            sb.append(m_data.get(i));
        }
        sb.append(")");
        return sb.toString();
    }

    private int countNumberOfBranches(Groupable[] lhss) throws JessException {

        for (int i=0; i<lhss.length; i++)
            lhss[i] = m_data.get(i).simplify();

        int product;
        if (m_name.equals(OR)) {
            product = 0;
            for (int i=0; i<lhss.length; i++) {
                product += lhss[i].getGroupSize();
            }
        }
        else {
            product = 1;
            for (int i=0; i<lhss.length; i++) {
                product *= lhss[i].getGroupSize();
            }
        }
        return product;
    }


    public Groupable simplify() throws JessException {

        Groupable[] lhss = new Groupable[m_data.size()];
        int product = countNumberOfBranches(lhss);

        if (product == 1) {
            Group g = new Group(OR);
            g.add(this);
            return g;
        }

        // Each element in rv is a new LHS being built.
        Group[] branches = new Group[product];

        for (int i=0; i<product; i++)
            branches[i] = new Group(AND);
        
        if (m_name.equals(OR)) {        
            // If we're an OR, then we're just the OR of all the children.
            // Each child in lhss is just a list of one or more OR-options
            int index=0;
            for (int i=0; i<lhss.length; i++)
                for (int j=0; j<lhss[i].getGroupSize(); j++)
                    branches[index++].add(lhss[i].getGroupable(j)); 
        }
        else {
            // Otherwise, it's a little more complex. We need to
            // distribute our OR children. "Repeat" is the number of
            // consecutive LHSs in the list a given sub-OR element
            // will be added to. It will get smaller and smaller and
            // end up as 1 for the last OR.

            int copies = product;
            for (int i=0; i<lhss.length; i++) {
                // If there's only one option in a branch, add it to every LHS.
                if (lhss[i].getGroupSize() == 1) {
                    for (int j=0; j<product; j++)
                        branches[j].add(lhss[i].getGroupable(0));
                }
            
                // If there are multiple options, we need to spread them
                // out.  We'll make "copies" copies of each option,
                // repeating the overall pattern "repeat" times. Each time
                // we encounter another OR in lhss, we divide "copies" by
                // the number of branches. If lhss looks like
                // 
                //  A C D 
                //  B   E
                //
                // Then the first and third entries are ORs, and we
                // want rv to look like
                //
                //  A A B B  copies == 2, repeat == 1
                //  C C C C
                //  D E D E  copies == 1, repeat == 2
            
                else {
                    copies /= lhss[i].getGroupSize();  
                    int repeat = product/(copies * lhss[i].getGroupSize());
                    int index=0;
                    for (int j=0; j<repeat; j++)
                        for (int k=0; k<lhss[i].getGroupSize(); k++)
                            for (int m=0; m<copies; m++)
                                branches[index++].add(lhss[i].getGroupable(k));
                }
            }
        }

        Group or = new Group(OR);
        for (int i=0; i<product; ++i) {
                
            if (branches[i].getGroupSize() == 1 &&
                !branches[i].getGroupable(0).getName().equals(NOT) &&
                !branches[i].getGroupable(0).getName().equals(TEST) &&
                !branches[i].getGroupable(0).getName().equals(EXISTS))
                or.add(branches[i].getGroupable(0));
            else
                or.add(branches[i]);
        }
        return or;
    }

    void insertInitialFacts() {
        Groupable g0 = getGroupable(0);
        if (m_name.equals(AND) && 
            ( m_data.size() == 0 ||
              (g0.getName().equals(LOGICAL) &&
               (isNegatedName(((Group) g0).getGroupable(0).getName()) ||
                ((Group) g0).getGroupable(0).getName().equals(TEST))) ||
              g0.getName().equals(NOT) ||
              g0.getName().equals(EXISTS) ||
              g0.getName().equals(TEST) ||
              g0.getBackwardChaining())) {
            m_data.addAtStart(g0 == null || !g0.getLogical() ?
                              s_initialFactPattern :
                              s_logicalInitialFactPattern);
        }            
    }

    // Call only on flattened LHSs with initial-facts added and
    // logical consistency checked.
    public void addToLHS(HasLHS l, Rete engine)
        throws JessException {

        for (int i=0; i<m_data.size(); i++) {
            if (m_data.get(i) instanceof Group)
                throw new JessException("Group.addToLHS",
                                        "Group detected:",
                                        "Called from illegal state");

            m_data.get(i).addToLHS(l, engine);
        }
    }

    public void flattenToGroup(Group g, int negcnt)
        throws JessException {
        if (m_name.equals(NOT))
            ++negcnt;
        else if (m_name.equals(EXISTS))
            negcnt += 2;
        else if (m_name.equals(OR))
            throw new JessException("Group.flattenToGroup",
                                    "OR detected:",
                                    "Illegal argument");
            
        insertInitialFacts();

        for (int i=0; i<m_data.size(); i++) {
            m_data.get(i).flattenToGroup(g, negcnt);            
        }
    }

    // To be called only on things that are already "flattened".

    public boolean isLogicallyConsistent() {
        for (int i=0; i<m_data.size(); ++i) {
            Pattern p1 = (Pattern) m_data.get(i);
            int n1 = p1.getNegated();
            for (int j=i+1; j<m_data.size(); ++j) {
                Pattern p2 = (Pattern) m_data.get(j);
                int n2 = p2.getNegated();
                if (p1.equals(p2) && n1 != n2)
                    return false;
            }
        }
        return true;
    }

    public int getGroupSize() {
        return m_data.size();
    }
    
    public Groupable getGroupable(int i) {
        return m_data.get(i);
    }  

    void setGroupable(Groupable g, int i) {
        m_data.set(g, i);
    }  

    public void addDirectlyMatchedVariables(Map map) throws JessException {
        for (int i=0; i<m_data.size(); i++) 
            m_data.get(i).addDirectlyMatchedVariables(map);
    }
    
    void renameVariables(Group g) throws JessException {
        HashMap map = new HashMap();
        HashMap subs = new HashMap();
        addDirectlyMatchedVariables(map);
        g.renameUnmentionedVariables(map, subs, m_data.size());
    }

    public void renameUnmentionedVariables(Map map, Map subs, int size)
        throws JessException {

        for (int i=0; i<m_data.size(); i++) {
            m_data.get(i).renameUnmentionedVariables(map, subs, size);
        }
    }

    private static class GroupableVector {
        private Groupable[] m_data = new Groupable[1];
        private int m_nData;
    
        void addAtStart(Groupable g) {
            if (m_data.length == m_nData) {
                Groupable[] temp = new Groupable[m_nData * 2];
                System.arraycopy(m_data, 0, temp, 0, m_nData);;
                m_data = temp;      
            }
            System.arraycopy(m_data, 0, m_data, 1, m_nData);
            m_data[0] = g;
            ++m_nData;
        }

        void add(Groupable g) {
      
            if (m_data.length == m_nData) {
                Groupable[] temp = new Groupable[m_nData * 2];
                System.arraycopy(m_data, 0, temp, 0, m_nData);;
                m_data = temp;      
            }
            m_data[m_nData++] = g;
        }
    
        Groupable get(int i) {
            return m_data[i];
        }


        void set(Groupable g, int i) {
            m_data[i] = g;
        }

        int size() {
            return m_nData;
        }
    }
}








